# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for storing data classes and complex type definitions that are used in the context of the
mlcvzoo_base.evaluation.geometric package
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Final, List, NamedTuple, Optional

import numpy as np

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.geometric_classifiction import GeometricClassification

# 3D List containing of the following:
#   1) One entry per data items / image
#   2) One entry per call of classes
#   3) One entry per annotation Annotations the matches the image (1) and the class (2)
# This is a list of BaseAnnotations in order to carry information about the image
# instead of just bounding boxes
EVALUATION_LIST_TYPE = List[List[List[BaseAnnotation]]]  # pylint: disable=invalid-name

CONFUSION_MATRIX_TYPE = List[List[int]]
CONFUSION_MATRIX_DICT_TYPE = Dict[str, Dict[str, int]]  # pylint: disable=invalid-name

DEFAULT_INT_VALUE: Final[int] = 0
DEFAULT_FLOAT_VALUE: Final[float] = 0.0


class EvaluationEntry(NamedTuple):
    image_path: str
    evaluation_objects: List[GeometricClassification]


# NOTE: Since this are the main Object Detection metrics, it's okay to have more instance
#       attributes
@dataclass
class GeometricMetrics:  # pylint: disable=too-many-instance-attributes
    """
    Dataclass for storing the main metrics that are computed for object detection algorithms
    """

    # TODO: rename to lower case, disable pylint only temporary!
    TP: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name
    FP: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name
    FN: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name
    PR: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    RC: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    F1: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    AP: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    COUNT: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name
    AVG_TP_IOU: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> GeometricMetrics:
        return GeometricMetrics(**input_dict)

    def __repr__(self):  # type: ignore
        return (
            f"TP: {self.TP}, "
            f"FP: {self.FP}, "
            f"FN: {self.FN}, "
            f"PR: {self.PR}, "
            f"RC: {self.RC}, "
            f"F1: {self.F1}, "
            f"AP: {self.AP}, "
            f"COUNT: {self.COUNT}"
            f"AVG_TP_IOU: {self.COUNT}"
        )

    def __str__(self):  # type: ignore
        return self.__repr__()


@dataclass
class MetricImageInfo:
    """
    Dataclass to store information about false positives and false negatives in the
    form of EvaluationEntry objects. It is used to have an exact relation between an image
    and the according false positive / false negative bounding boxes. The ground
    truth data is added to be able to visualize the expected bounding boxes.
    """

    gt_evaluation_entry: Optional[EvaluationEntry] = None
    tp_evaluation_entry: Optional[EvaluationEntry] = None
    fn_evaluation_entry: Optional[EvaluationEntry] = None
    fp_evaluation_entry: Optional[EvaluationEntry] = None
    fn_matched_fp_evaluation_entry: Optional[EvaluationEntry] = None


# 1st key: Class Identifier as string
# 2nd key: Image Path
# 2nd value: The MetricImageInfo for this class name and image path
METRIC_IMAGE_INFO_TYPE = Dict[str, Dict[str, MetricImageInfo]]  # pylint: disable=invalid-name


# 1st key: iou-threshold
# 2nd key: Type of the size of the bounding-box => Any of GeometricSizeTypes.BBOX_SIZE_TYPE
# 3rd key: Class Identifier as string
# value: The computed metrics of type GeometricMetrics
METRIC_DICT_TYPE = Dict[
    float, Dict[str, Dict[str, GeometricMetrics]]
]  # pylint: disable=invalid-name


def build_metric_dict_from_dict(
    input_dict: Dict[str, Dict[str, Dict[str, Dict[str, Any]]]]
) -> METRIC_DICT_TYPE:
    metric_dict: METRIC_DICT_TYPE = {}

    for key_0, value_0 in input_dict.items():
        metric_dict[float(key_0)] = {}
        for key_1, value_1 in value_0.items():
            metric_dict[float(key_0)][key_1] = {}
            for key_2, od_metrics_dict in value_1.items():
                metric_dict[float(key_0)][key_1][key_2] = GeometricMetrics.from_dict(
                    input_dict=od_metrics_dict
                )

    return metric_dict


@dataclass
class GeometricEvaluationMetrics:
    """
    Dataclass for storing the output of an object detection evaluation.
    The metrics_dict stores the actual computed metrics, while the metrics_image_info_dict
    stores debugging information to be able to analyze false positives and false negatives.

    The model_specifier indicates for which model the metrics have been computed.
    """

    model_specifier: str
    metrics_dict: METRIC_DICT_TYPE = field(default_factory=lambda: {})
    metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE = field(default_factory=lambda: {})


@dataclass
class GeometricEvaluationComputingData:
    """
    Dataclass for storing data structures that are needed to computed object detection metrics
    """

    # 1st key: Type of the size of the bounding-box => Any of GeometricSizeTypes.BBOX_SIZE_TYPE
    # 2nd key: Class Identifier as string
    # value: The number of ground truth boxes for the combination of keys
    gt_counter_dict: Dict[str, Dict[str, int]] = field(default_factory=lambda: {})

    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of GeometricSizeTypes.BBOX_SIZE_TYPE
    # Cumulative array indicating the false positives of the dataset
    false_positives_dict: Dict[float, Dict[str, np.ndarray]] = field(  # type: ignore[type-arg]
        default_factory=lambda: {}
    )

    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of GeometricSizeTypes.BBOX_SIZE_TYPE
    # Cumulative array indicating the true positives of the dataset
    true_positives_dict: Dict[float, Dict[str, np.ndarray]] = field(  # type: ignore[type-arg]
        default_factory=lambda: {}
    )

    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of GeometricSizeTypes.BBOX_SIZE_TYPE
    # Cumulative array indicating the iou value of an evaluation. This can be used to determine
    # false positives and false negatives, as well as to see the 'real' geometric overlap of
    # an prediction.
    iou_dict: Dict[str, np.ndarray] = field(default_factory=lambda: {})  # type: ignore[type-arg]

    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of GeometricSizeTypes.BBOX_SIZE_TYPE
    # value: Array indicating the score for each data item
    scores: Dict[float, Dict[str, np.ndarray]] = field(  # type: ignore[type-arg]
        default_factory=lambda: {}
    )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for different utility operations regarding drawing on images"""

import colorsys
import copy
import math
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
from PIL import Image, ImageDraw, ImageFont

from mlcvzoo_base.api.data.annotation_attributes import AnnotationAttributes
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.geometric_classifiction import GeometricClassification
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.api.data.types import FrameShape, ImageType, Point2f, PolygonTypeNP


def determine_draw_parameters(
    frame_shape: FrameShape, thickness_factor: float, font_scale_factor: float
) -> Tuple[int, float]:
    """
    Determines parameters for drawing. Thickness of lines and font are scaled by the image size.

    Args:
        frame_shape: Shape of given frame.
        thickness_factor: Factor for line thickness.
        font_scale_factor: Factor for font size.

    Returns:
        Line thickness and font scale.

    """
    thickness = math.ceil(min(frame_shape.width, frame_shape.height) * thickness_factor)
    font_scale = min(frame_shape.width, frame_shape.height) * font_scale_factor

    return thickness, font_scale


def generate_detector_colors(num_classes: int):  # type: ignore
    """Generates a color palette for object detector.

    Args:
        num_classes: int, the number of classes to distinguish

    Returns:
        List[Tuples(int, int, int)] specifying bounding box color
        with len(List) == number of classes

    """

    # Generate colors for drawing bounding boxes.
    hsv_tuples = [(x / num_classes, 1.0, 1.0) for x in range(num_classes)]

    hsv_colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))

    rgb_colors = list(
        map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), hsv_colors)
    )

    np.random.seed(10101)  # Fixed seed for consistent colors across runs.
    np.random.shuffle(rgb_colors)  # Shuffle colors to decorrelate adjacent classes.
    np.random.seed(None)  # Reset seed to default.

    return rgb_colors


def draw_on_image(
    frame: ImageType,
    rgb_colors: List[Tuple[int, int, int]],
    cv2_font: int = cv2.FONT_HERSHEY_COMPLEX,
    bounding_boxes: Optional[List[BoundingBox]] = None,
    segmentations: Optional[List[Segmentation]] = None,
    predictions: Optional[List[GeometricClassification]] = None,
    flip_image: bool = False,
    draw_caption: Optional[bool] = True,
    thickness: Optional[int] = None,
    font_scale: float = 0.9,
    draw_score: Optional[bool] = True,
    include_content_in_caption: bool = False,
    classifications: Optional[List[Classification]] = None,
    auto_scale: bool = True,
    font_scale_factor: float = 0.0015,
    thickness_factor: float = 0.0015,
) -> ImageType:
    """Draws annotations on the given image

    Args:
        frame: Numpy array, the image
        rgb_colors: List of colors, for matching a class-name to an index of the rgb-color list
        cv2_font: int, font of cv2 textual annotations
        bounding_boxes: Optional List of BoundingBox objects
        segmentations: Optional List of Segmentation objects
        predictions: Optional List of GeometricClassification objects
        flip_image: Bool, whether to flip the image or not
        draw_caption: Optional bool, whether to draw a caption text or not
        thickness: Optional int, defines the thickness of the annotation lines
        font_scale: float, defines the size of a font
        draw_score: Whether to draw the score of the bounding boxes or segmentations
        include_content_in_caption: Whether to add the content of an object to the caption
        classifications: Optional List of Classification objects
        auto_scale: Whether to automatically scale the text and thickness.
        font_scale_factor: Factor for automatic scaling of the font size
        thickness_factor: Factor for automatic scaling of the text thickness.

    Returns:
        Numpy array, the given image decorated with annotations
    """

    frame = copy.deepcopy(frame)

    _predictions: List[GeometricClassification] = []

    if bounding_boxes is not None:
        _predictions.extend(bounding_boxes)

    if segmentations is not None:
        _predictions.extend(segmentations)

    if predictions is not None:
        _predictions.extend(predictions)

    for prediction in _predictions:
        if (
            include_content_in_caption
            and isinstance(prediction, (GeometricClassification, AnnotationAttributes))
            and prediction.content != ""  # type: ignore[attr-defined]
        ):
            label = f"{prediction.class_name}__{prediction.content}"  # type: ignore[attr-defined]
        else:
            label = prediction.class_name

        frame = draw_polygon_cv2(
            frame=frame,
            polygon=prediction.polygon(),
            color=rgb_colors[prediction.class_id],
            flip_image=flip_image,
            cv2_font=cv2_font,
            label=label,
            draw_caption=draw_caption,
            thickness=thickness,
            font_scale=font_scale,
            score=prediction.score if draw_score else None,
            auto_scale=auto_scale,
            font_scale_factor=font_scale_factor,
            thickness_factor=thickness_factor,
        )

    if classifications is not None:
        frame = draw_classification_cv2(
            frame=frame, classifications=classifications, cv2_font=cv2_font
        )

    return frame


def draw_classification_cv2(
    frame: ImageType,
    classifications: List[Classification],
    cv2_font: int = cv2.FONT_HERSHEY_DUPLEX,
    thickness: int = 1,
    font_scale: float = 0.9,
    auto_scale: bool = True,
    font_scale_factor: float = 0.0015,
    thickness_factor: float = 0.0015,
) -> ImageType:
    """
    Draws classification text on the given image using OpenCV.

    Args:
        frame (ImageType): The image on which to draw the classification.
        classifications (List[Classification]): List of classification objects.
        cv2_font (int, optional): Font type for the text.
        thickness (int, optional): Thickness of the text. Defaults to 1.
        font_scale (float, optional): Scale factor for the font size.
        auto_scale (bool, optional): Whether to automatically scale the text and thickness.
        font_scale_factor (float, optional): Factor for automatic scaling of the font size
        thickness_factor (float, optional): Factor for automatic scaling of the text thickness.

    Returns:
        ImageType: The image with the classification text drawn on it.
    """

    if not classifications:
        return frame

    max_index = np.argmax([c.score for c in classifications])
    max_classification = classifications[max_index]

    text = "Class: {} - {:.2f}".format(
        str(max_classification.class_identifier).strip(), max_classification.score
    )

    frame_shape = FrameShape(*frame.shape)

    if auto_scale:
        thickness, font_scale = determine_draw_parameters(
            frame_shape=frame_shape,
            thickness_factor=thickness_factor,
            font_scale_factor=font_scale_factor,
        )

    text_size = cv2.getTextSize(text, fontFace=cv2_font, fontScale=font_scale, thickness=thickness)

    # Automatically determine the split of the given text in different lines. This depends
    # on the width of the image, the size of the text and a buffer.
    text_lines = []
    _text_line = ""
    for index, character in enumerate(text):
        if index % (int(len(text) * frame_shape.width / text_size[0][0]) - 5) == 0 and index != 0:
            text_lines.append(_text_line)
            _text_line = ""

        _text_line += character
    text_lines.append(_text_line)

    # Create a white border at the top of the image, where the text of
    # the classification output is put in.
    border_size_constant = 1.8
    border_size = int(text_size[0][1] * border_size_constant * len(text_lines))
    mean = 255
    frame = cv2.copyMakeBorder(
        src=frame,
        top=border_size,
        bottom=0,
        left=0,
        right=0,
        borderType=cv2.BORDER_CONSTANT,
        value=[mean, mean, mean],
    )

    # The parameter for the position of the text are manually optimized to fit the text
    # into the created image border.
    draw_text_constant = 1.3
    draw_text_x_origin = 5
    # Draw the classification output in the created image border
    for index, text_line in enumerate(text_lines):
        text_line = text_line.rstrip()
        text_line = text_line.lstrip()

        cv2.putText(
            img=frame,
            text=text_line,
            # Set the y-orign for each line of text to be drawn
            org=(draw_text_x_origin, int(text_size[0][1] * draw_text_constant) * (index + 1)),
            fontFace=cv2_font,
            fontScale=font_scale,
            color=(0, 0, 0),
            thickness=thickness,
        )

    return frame


def draw_polygon_cv2(
    frame: ImageType,
    color: Tuple[int, int, int],
    polygon: Union[PolygonTypeNP, List[Point2f]],
    flip_image: bool = False,
    cv2_font: int = cv2.FONT_HERSHEY_COMPLEX,
    label: Optional[str] = None,
    draw_caption: Optional[bool] = True,
    thickness: Optional[int] = None,
    font_scale: float = 0.9,
    score: Optional[float] = None,
    auto_scale: bool = True,
    font_scale_factor: float = 0.0015,
    thickness_factor: float = 0.0015,
) -> ImageType:
    """Draw a polygon on a given image

    Args:
        frame: Numpy array, the image
        color: Tuple of three int values, a rgb color
        polygon: List of Tuples, the polygon coordinates
        flip_image: Bool, whether to flip the image or not
        cv2_font: int, font of cv2 textual annotations
        label: Optional string, label (class information) of polygon
        draw_caption: Optional bool, whether to draw a caption text or not
        thickness: Optional int, defines the thickness of the polygon lines
        font_scale: float, defines the size of a font
        score: Optional float, defines the confidence for the label of the polygon
        auto_scale: Whether to use the automatic scaling of the thickness and font scale
        font_scale_factor: Factor for automatic scaling the font scale of the text in the image
        thickness_factor: Factor for automatic scaling of thickness of the text size in the image

    Returns:
        Numpy array, the given image decorated with a polygon
    """

    if polygon is None:
        return frame

    if thickness is None:
        thickness = (frame.shape[0] + frame.shape[1]) // 450

    font_thickness = thickness

    frame_shape = FrameShape(*frame.shape)
    if auto_scale:
        font_thickness, font_scale = determine_draw_parameters(
            frame_shape=frame_shape,
            thickness_factor=thickness_factor,
            font_scale_factor=font_scale_factor,
        )
        thickness = int(font_thickness) + 1

    if flip_image:
        # TODO: what to do when image is flipped?
        pass
    else:
        # Apply same behavior as cv2.rectangle for thickness.
        # It is the thickness of the rectangle border line in px.
        # Thickness of -1 px will fill the rectangle shape by the
        # specified color.
        pts = np.asarray(polygon, np.int32)  # type: ignore[var-annotated]
        pts = pts.reshape((-1, 1, 2))
        if thickness == -1:
            frame = cv2.fillPoly(
                img=frame,
                pts=[pts],
                color=color,
            )
        else:
            frame = cv2.polylines(
                img=frame, pts=[pts], isClosed=True, color=color, thickness=thickness
            )
    if draw_caption and label is not None:
        if score is not None:
            caption = f"{label}_{score:.2f}"
        else:
            caption = label

        cv2.putText(
            img=frame,
            text=caption,
            org=(int(polygon[0][0]), int(polygon[0][1]) - int(0.0155 * frame_shape.height)),
            fontFace=cv2_font,
            fontScale=font_scale,
            color=color,
            thickness=font_thickness,
        )

    return frame


def draw_bbox_cv2(
    frame: ImageType,
    color: Tuple[int, int, int],
    box: Box,
    flip_image: bool = False,
    cv2_font: int = cv2.FONT_HERSHEY_COMPLEX,
    label: Optional[str] = None,
    draw_caption: Optional[bool] = True,
    thickness: Optional[int] = None,
    font_scale: float = 0.9,
    score: Optional[float] = None,
    text_origin: Optional[Tuple[int, int]] = None,
) -> ImageType:
    """Draws a bounding box on the given image

    Args:
        frame: Numpy array, the image
        color: Tuple of three int values, a rgb color
        box: Box, a bounding box object
        flip_image: Bool, whether to flip the image or not
        cv2_font: int, font of cv2 textual annotations
        label: Optional string, label (class information) of polygon
        draw_caption: Optional bool, whether to draw a caption text or not
        thickness: Optional int, defines the thickness of the box lines
        font_scale: float, defines the size of a font
        score: Optional float, defines the confidence for the label of the polygon
        text_origin: position of text in image

    Returns:
        Numpy array, the given image decorated with a bounding box
    """

    img_w = frame.shape[0]

    if thickness is None:
        thickness = (frame.shape[0] + frame.shape[1]) // 450

    if flip_image:
        frame = cv2.rectangle(
            frame,
            (img_w - box.xmin, box.ymin),
            (img_w - box.xmax, box.ymax),
            color,
            thickness,
        )
    else:
        # Apply same behavior as cv2.rectangle for thickness.
        # It is the thickness of the rectangle border line in px.
        # Thickness of -1 px will fill the rectangle shape by the
        # specified color.
        pts = box.polygon().astype(np.int_)  # type: ignore[var-annotated]
        if thickness == -1:
            frame = cv2.fillPoly(
                img=frame,
                pts=[pts],
                color=color,
            )
        else:
            frame = cv2.polylines(
                img=frame,
                pts=[pts],
                isClosed=True,
                color=color,
                thickness=thickness,
            )

    if draw_caption and label is not None:
        if text_origin is None:
            text_x = int(box.top_left[0])
            text_y = max(int(box.top_left[1]) - 5, 20)
            text_origin = (text_x, text_y)

        if score is not None:
            caption = f"{label} - {score:.2f}"
        else:
            caption = label

        cv2.putText(
            img=frame,
            text=caption,
            org=text_origin,
            fontFace=cv2_font,
            fontScale=font_scale,
            color=color,
            thickness=thickness,
        )

    return frame


def draw_on_pil_image(
    image: Union[ImageType, Image.Image],
    bounding_boxes: List[BoundingBox],
    font_path: str,
    rgb_colors: List[Tuple[int, int, int]],
    thickness: int = 2,
    fill_background: bool = True,
    is_ground_truth: bool = False,
) -> ImageType:
    """Draw a given list of bounding-boxes on a given image using pillow.

    Args:
        image: either a Numpy array or a PIL Image object
        bounding_boxes: List of BoundingBox objects
        font_path: String, path to the font file which should be used
        rgb_colors: List of colors, for matching a class-name to an index of the rgb-color list
        thickness: int, thickness of the box lines
        fill_background: Bool, whether to put the label on white background or not
        is_ground_truth: Bool, depending on ground truth or not, the label will be drawn
            in the upper left corner (no gt) or in the lower left corner (is gt)

    Returns:
        Numpy array, the given image decorated with annotations
    """

    white_color = (255, 255, 255, 15)

    if type(image) is np.ndarray:
        image = Image.fromarray(image)

    assert isinstance(image, Image.Image)

    font = ImageFont.truetype(
        font=font_path, size=np.floor(0.042 * image.size[1] + 0.1).astype("int32")
    )

    draw = ImageDraw.Draw(image)

    for bounding_box in bounding_boxes:
        label = "{} {:.2f}".format(bounding_box.class_name, bounding_box.score)

        # the anchor xy can be zero because we are only interested in the necessary
        # text width and height
        label_box = draw.textbbox(xy=(0, 0), text=label, font=font)
        label_size = (label_box[2], label_box[3])

        if is_ground_truth:
            text_origin = (
                float(bounding_box.box().xmin),
                float(bounding_box.box().ymax - label_size[1]),
            )
        else:
            if bounding_box.box().ymin - label_size[1] >= 0:
                text_origin = (
                    float(bounding_box.box().xmin),
                    float(bounding_box.box().ymin - label_size[1]),
                )
            else:
                text_origin = (
                    float(bounding_box.box().xmin),
                    float(bounding_box.box().ymin + 1),
                )

        color = rgb_colors[bounding_box.class_id]

        if fill_background:
            draw.rectangle(
                (
                    text_origin,
                    (text_origin[0] + label_size[0], text_origin[1] + label_size[1]),
                ),
                fill=white_color,
            )

            draw.text(text_origin, label, fill=(0, 0, 0, 15), font=font)
        else:
            draw.text(text_origin, label, fill=color, font=font)

        for draw_step in range(thickness):
            draw.rectangle(
                (
                    float(bounding_box.box().xmin + draw_step),
                    float(bounding_box.box().ymin + draw_step),
                    float(bounding_box.box().xmax - draw_step),
                    float(bounding_box.box().ymax - draw_step),
                ),
                outline=color,
            )

    del draw

    return np.asarray(image)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for evaluating currently trained model checkpoints"""

import logging
from typing import List, Union

from tqdm import tqdm

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.geometric_classifiction import GeometricClassification
from mlcvzoo_base.api.model import (
    ConfigurationType,
    DataType,
    ObjectDetectionModel,
    SegmentationModel,
)
from mlcvzoo_base.evaluation.geometric.data_classes import GeometricEvaluationMetrics
from mlcvzoo_base.evaluation.geometric.metrics_computation import (
    EvaluationContexts,
    MetricsComputation,
)

logger = logging.getLogger(__name__)


def evaluate_with_model(
    gt_annotations: List[BaseAnnotation],
    iou_thresholds: List[float],
    model: Union[
        ObjectDetectionModel[ConfigurationType, DataType],
        SegmentationModel[ConfigurationType, DataType],
    ],
    evaluation_context: str = EvaluationContexts.ALL.value,
) -> GeometricEvaluationMetrics:
    """Compute the metric for the given object detection model. The evaluation is performed
    on the basis of the given ground truth annotations.

    Args:
        model: The model that should be evaluated
        gt_annotations: ground truth annotations where the model should be evaluated on
        iou_thresholds: List of thresholds for which a metrics should be computed
        evaluation_context: Context of the evaluation

    Returns:
        The computed object detection metrics for this model
    """

    predictions_list: List[List[GeometricClassification]] = []

    process_bar = tqdm(
        gt_annotations,
        desc=f"Run prediction",
    )

    for index, gt_annotation in enumerate(process_bar):
        # Every ObjectDetectionModel and SegmentationModel returns GeometricClassifications
        _, predictions = model.predict(data_item=gt_annotation.image_path)
        predictions_list.append(predictions)  # type: ignore[arg-type]

    return MetricsComputation(
        model_specifier=model.unique_name,
        iou_thresholds=iou_thresholds,
        gt_annotations=gt_annotations,
        predictions_list=predictions_list,
        mapper=model.mapper,
        evaluation_context=evaluation_context,
    ).compute_metrics()


def evaluate_with_precomputed_data(
    model_specifier: str,
    gt_annotations: List[BaseAnnotation],
    iou_thresholds: List[float],
    predictions_list: List[List[GeometricClassification]],
    mapper: AnnotationClassMapper,
    evaluation_context: str = EvaluationContexts.ALL,
) -> GeometricEvaluationMetrics:
    """
    Compute the object detection metrics taking precomputed (predicted) bounding boxes and
    ground truth annotations.

    IMPORTANT REMARK: The index of the lists 'ground_truth_annotations'
                      and 'predictions_list' have to match exactly. This means
                      index 0 indicates the ground truth data and predicted bounding boxes
                      for image 0.

    Args:
        model_specifier: A string to indicate with which model the precomputed bounding boxes
                         have been predicted
        gt_annotations: The ground truth data as basis for the evaluation
        iou_thresholds: List of thresholds for which a metrics should be computed
        predictions_list: A list containing the predicted GeometricClassifications for each
                          image of ground truth data
       mapper: An AnnotationClassMapper object that states the mapping of Class-IDs/Class-Names
               to ClassIdentifier(s)
       evaluation_context: Context of the evaluation

    Returns:
        The computed object detection metrics
    """

    return MetricsComputation(
        model_specifier=model_specifier,
        iou_thresholds=iou_thresholds,
        gt_annotations=gt_annotations,
        predictions_list=predictions_list,
        mapper=mapper,
        evaluation_context=evaluation_context,
    ).compute_metrics()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

""" Class for Bounding Box Annotation"""

from __future__ import annotations

import math
from typing import Any, Dict, List, Optional

from mlcvzoo_base.api.data.annotation_attributes import AnnotationAttributes
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.geometric_classifiction import GeometricClassification
from mlcvzoo_base.api.data.types import PolygonTypeNP, polygon_equal
from mlcvzoo_base.api.structs import float_equality_precision


class BoundingBox(GeometricClassification, AnnotationAttributes):
    """
    A class for defining the data object consumed by ObjectDetection models.
    It is mainly described by the box attribute, which covers an rectangular
    area of an image and is associated with a certain class
    """

    def __init__(
        self,
        box: Box,
        class_identifier: ClassIdentifier,
        score: float,
        difficult: bool = False,
        occluded: bool = False,
        content: str = "",
        model_class_identifier: Optional[ClassIdentifier] = None,
        background: bool = False,
        meta_attributes: Optional[Dict[Any, Any]] = None,
    ):
        Classification.__init__(
            self,
            class_identifier=class_identifier,
            model_class_identifier=model_class_identifier,
            score=score,
        )
        AnnotationAttributes.__init__(
            self,
            difficult=difficult,
            occluded=occluded,
            content=content,
            background=background,
            meta_attributes=meta_attributes,
        )
        self.__box = box
        # The polygon is created lazily
        self.__polygon: Optional[PolygonTypeNP] = None

    def box(self) -> Box:
        return self.__box

    def ortho_box(self) -> Box:
        if self.__box.is_orthogonal():
            return self.__box

        return self.__box.ortho_box()

    def polygon(self) -> PolygonTypeNP:
        if self.__polygon is None:
            self.__polygon = self.__box.polygon()

        return self.__polygon

    def __eq__(self, other: object) -> bool:
        try:
            # NOTE: Since floats may very for different systems, don't check the score for equality,
            #       but allow it to be in a reasonable range
            return (
                self.box() == other.box()  # type: ignore[attr-defined]
                and self.class_identifier.class_id == other.class_identifier.class_id  # type: ignore[attr-defined]
                and self.class_identifier.class_name == other.class_identifier.class_name  # type: ignore[attr-defined]
                and self.model_class_identifier.class_id
                == other.model_class_identifier.class_id  # type: ignore[attr-defined]
                and self.model_class_identifier.class_name
                == other.model_class_identifier.class_name  # type: ignore[attr-defined]
                and self.occluded == other.occluded  # type: ignore[attr-defined]
                and self.difficult == other.difficult  # type: ignore[attr-defined]
                and self.background == other.background  # type: ignore[attr-defined]
                and self.content == other.content  # type: ignore[attr-defined]
                and self.meta_attributes == other.meta_attributes  # type: ignore[attr-defined]
                and math.isclose(a=self.score, b=other.score, abs_tol=float_equality_precision)  # type: ignore[attr-defined]
                and polygon_equal(polygon=self.polygon(), other_polygon=other.polygon())  # type: ignore[attr-defined]
            )
        except AttributeError:
            return False

    def __str__(self):  # type: ignore
        return (
            f"BoundingBox("
            f'class_identifier="{str(self.class_identifier)}", '
            f'model_class_identifier="{str(self.model_class_identifier)}", '
            f"box={self.__box}, "
            f"score={self.score}, "
            f"difficult={self.difficult}, "
            f"occluded={self.occluded}, "
            f"background={self.background}, "
            f'content="{self.content}", '
            f"meta_attributes={self.meta_attributes}"
            f")"
        )

    def __repr__(self):  # type: ignore
        return (
            f"BoundingBox("
            f"class_identifier={repr(self.class_identifier)}, "
            f"model_class_identifier={repr(self.model_class_identifier)}, "
            f"box={self.__box}, "
            f"score={self.score}, "
            f"difficult={self.difficult}, "
            f"occluded={self.occluded}, "
            f"background={self.background}, "
            f'content="{self.content}", '
            f"meta_attributes={self.meta_attributes}"
            f")"
        )

    def to_dict(self, raw_type: bool = False, reduced: bool = False) -> Dict[str, Any]:
        """
        Generate a dictionary representation of the object. The raw_type defined whether
        all data attributes should be converted to python native data types or should
        be kept as is. The reduced parameter allows to create a more compact representation
        where only the main attributes of the object are present.

        Args:
            raw_type: Whether to convert the data attributes to python native data types
            reduced: Whether to create a more compact representation

        Returns:
            The created dictionary representation
        """
        if reduced:
            return {
                "box": self.__box if raw_type else self.__box.to_dict(),
                "class_id": self.class_id,
                "class_name": self.class_name,
                "model_class_id": self.model_class_identifier.class_id,
                "model_class_name": self.model_class_identifier.class_name,
                "score": self.score,
            }

        return {
            "box": self.__box if raw_type else self.__box.to_dict(),
            "class_identifier": (
                self.class_identifier if raw_type else self.class_identifier.to_dict()
            ),
            "model_class_identifier": (
                self.model_class_identifier if raw_type else self.model_class_identifier.to_dict()
            ),
            "score": self.score,
            "difficult": self.difficult,
            "occluded": self.occluded,
            "background": self.background,
            "content": self.content,
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any], reduced: bool = False) -> BoundingBox:
        # fmt: off
        if reduced:
            return BoundingBox(**{
                "box": Box(**input_dict["box"]),
                "class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["class_id"],
                    "class_name": input_dict["class_name"],
                }),
                "model_class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["model_class_id"],
                    "class_name": input_dict["model_class_name"],
                }),
                "score": input_dict["score"],
                "difficult": False,
                "occluded": False,
                "background": False,
                "content": "",
            })
        else:
            return BoundingBox(**{
                "box": Box(**input_dict["box"]),
                "class_identifier": ClassIdentifier(
                    **input_dict["class_identifier"]
                ),
                "model_class_identifier": ClassIdentifier(
                    **input_dict["model_class_identifier"]
                ),
                "score": input_dict["score"],
                "difficult": input_dict["difficult"],
                "occluded": input_dict["occluded"],
                "background": input_dict["background"],
                "content": input_dict["content"],
            })
        # fmt: on

    def to_json(self) -> Any:
        return self.to_dict(raw_type=False)

    def to_list(self) -> List[int]:
        """Transforms the BoundingBox object to a list of its coordinates.

        REMARK: This function uses the ortho_box() since is only defined for
                orthogonal bounding boxes

        Returns:
            A 1x4 list of the objects coordinates [xmin, ymin, xmax, ymax]
        """
        return self.ortho_box().to_list()

    def copy_bounding_box(self, class_identifier: ClassIdentifier) -> BoundingBox:
        """Constructs a copy of the BoundingBox object with the given class-identifier.

        Args:
            class_identifier: The class-identifier of the new BoundingBox object

        Returns:
            The created copy
        """
        return BoundingBox(
            box=self.__box,
            class_identifier=class_identifier,
            score=self.score,
            difficult=self.difficult,
            occluded=self.occluded,
            background=self.background,
            content=self.content,
            model_class_identifier=self.model_class_identifier,
        )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for parsing Label Studio formatted annotations of single json files that are
generated by Label Studio per image
"""
import copy
import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, cast

import cv2
import numpy as np
from attr import define

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box, rotate_point
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.api.data.types import FrameShape, Point2f
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError, ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerLabelStudioInputDataConfig,
)
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.utils.file_utils import get_image_file_dict

logger = logging.getLogger(__name__)


@define
class LabelStudioAnnotationData:
    image_shape: Optional[Tuple[int, int]] = None
    classifications: List[Classification] = []
    bounding_boxes: List[BoundingBox] = []
    segmentations: List[Segmentation] = []


def __merge_single_entry(results: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Merge the annotation data of a label studio result entry.
    Typically, label studio creates multiple different entries
    per result type. To parse the annotation content into a
    mlcvzoo annotation we want all the different entries to be
    merged into one result entry that contains all the necessary
    information.

    Args:
        results: The results that should be merged

    Returns:
        The merged results
    """
    merged_results = {}
    for result in results:
        result_id = result["id"]
        result_type = result["type"]
        if result_id not in merged_results:
            merged_results[result_id] = result
            merged_results[result_id]["type"] = [result_type]
            continue

        merged_results[result_id]["value"].update(result["value"])
        merged_results[result_id]["type"] += [result_type]

    return merged_results


def get_label_studio_annotation_dict(
    label_studio_annotation_path: str,
) -> Optional[Dict[Any, Any]]:
    """
    Read the content of a label studio annotation file in json format.
    These files are generated by label studio per annotation task (image).

    If the given label_studio_annotation_path exists, the json data will
    be parsed and merged as preparation for the main parsing methods
    that will translate the data into mlcvzoo annotations.

    Args:
        label_studio_annotation_path: The path to the annotation file

    Returns:
        The dictionary with the parsed annotation data
    """
    with open(label_studio_annotation_path, "r") as label_studio_file:
        try:
            annotation_data = cast(Dict[Any, Any], json.load(label_studio_file))

            annotation_data_orig = copy.deepcopy(annotation_data)

            try:
                # list() is needed for correct type conversion from dict_values to list
                annotation_data["result"] = list(
                    __merge_single_entry(annotation_data["result"]).values()
                )
                return annotation_data
            except KeyError as error:
                # This is not an exception that is breaking the code.
                # We can still use the original parsed data to continue with.
                logger.warning(
                    "Could not merge data due to error=%s, "
                    "will continue with original data" % repr(error)
                )
                return annotation_data_orig

        except (KeyError, json.decoder.JSONDecodeError) as error:
            logger.exception(
                msg="Could not parse annotation file '%s'. Annotation will be skipped."
                % label_studio_annotation_path,
                exc_info=error,
            )

    return None


def get_label_studio_annotation_dict_single_file(
    label_studio_annotation_path: str,
) -> Optional[List[Dict[str, Any]]]:
    """
    Read the content of a label studio annotation file in json format.
    This is expected to be an annotation file that has been exported
    via the label studio UI and contains annotations of multiple annotation tasks.

    If the given label_studio_annotation_path exists, the json data will
    be parsed and merged as preparation for the main parsing methods
    that will translate the data into mlcvzoo annotations.

    Args:
        label_studio_annotation_path: The path to the annotation file

    Returns:
        The dictionary with the parsed annotation data
    """

    with open(label_studio_annotation_path, "r") as label_studio_file:
        try:
            annotation_data = cast(List[Dict[str, Any]], json.load(label_studio_file))

            annotation_data_orig = copy.deepcopy(annotation_data)

            try:
                for index, project in enumerate(annotation_data):
                    annotations = project["annotations"]

                    for annotation_index, annotation in enumerate(annotations):
                        # list() is needed for correct type conversion from dict_values to list
                        annotation_data[index]["annotations"][annotation_index]["result"] = list(
                            __merge_single_entry(
                                annotation_data[index]["annotations"][annotation_index]["result"]
                            ).values()
                        )
            except KeyError as error:
                # This is not an exception that is breaking the code.
                # We can still use the original parsed data to continue with.
                logger.warning(
                    "Could not merge data due to error=%s, "
                    "will continue with original data" % repr(error)
                )
                return annotation_data_orig

            return annotation_data

        except json.decoder.JSONDecodeError as json_error:
            logger.exception(
                msg="Could not parse annotation file '%s'. Annotation will be skipped."
                % label_studio_annotation_path,
                exc_info=json_error,
            )

    return None


def parse_classification(
    mapper: AnnotationClassMapper,
    label_studio_value: Dict[str, Any],
) -> Optional[Classification]:
    """
    Try to parse a single Classification object from a given label studio result value (one
    entry of a result in the annotation file).

    Args:
        mapper: The mapper that should be used to translate class names
        label_studio_value: The input value for parsing the BoundingBox

    Returns:
        Optional[Classification]: The parsed Classification object
    """
    choices = label_studio_value.get("choices", None)
    if choices:
        # Currently we only support single class training, therefore take the first
        # choice occurrence
        label_studio_class_name = choices[0]
    else:
        # We don't accept annotations without a class
        return None

    try:
        class_name = mapper.map_annotation_class_name_to_model_class_name(
            class_name=label_studio_class_name
        )
        class_id = mapper.map_annotation_class_name_to_model_class_id(
            class_name=label_studio_class_name
        )
    except ClassMappingNotFoundError:
        logger.debug(
            "Could not find a valid class-mapping for class-name '%s'. "
            "Classification will be skipped",
            label_studio_class_name,
        )
        return None

    return Classification(
        class_identifier=ClassIdentifier(class_id=class_id, class_name=class_name),
        score=1.0,
    )


def parse_bounding_box(
    mapper: AnnotationClassMapper, label_studio_value: Dict[str, Any], image_shape: Tuple[int, int]
) -> Optional[BoundingBox]:
    """
    Try to parse a single BoundingBox object from a given label studio result value (one
    entry of a result in the annotation file).

    Args:
        mapper: The mapper that should be used to translate class names
        label_studio_value: The input value for parsing the BoundingBox
        image_shape: The shape of the related image

    Returns:
        Optional[BoundingBox]: The parsed BoundingBox object
    """
    rectanglelabels = label_studio_value.get("rectanglelabels", None)
    labels = label_studio_value.get("labels", None)
    if rectanglelabels:
        label_studio_class_name = rectanglelabels[0]
    elif labels:
        label_studio_class_name = labels[0]
    else:
        # We don't except annotations without a class
        return None

    try:
        class_name = mapper.map_annotation_class_name_to_model_class_name(
            class_name=label_studio_class_name
        )
        class_id = mapper.map_annotation_class_name_to_model_class_id(
            class_name=label_studio_class_name
        )
    except ClassMappingNotFoundError:
        logger.debug(
            "Could not find a valid class-mapping for class-name '%s'. "
            "BoundingBox will be skipped",
            label_studio_class_name,
        )
        return None

    # LabelStudio has defined its rotation around the top-left point of the bounding box.
    # The RotatedBox of the MLCVZoo uses the opencv definition and rotates around the center
    # point of the bounding box. Therefore, we have to transform the LabelStudio definition
    # to the opencv definition.

    top_left_x = float(image_shape[1] * label_studio_value["x"] / 100)
    top_left_y = float(image_shape[0] * label_studio_value["y"] / 100)

    dim_x = float(image_shape[1] * label_studio_value["width"] / 100)
    dim_y = float(image_shape[0] * label_studio_value["height"] / 100)

    rotation = label_studio_value["rotation"]

    content = label_studio_value.get("text", [""])[0]

    if rotation > 0.0:
        top_left: Point2f = (top_left_x, top_left_y)
        top_right: Point2f = (top_left_x + dim_x, top_left_y)
        bottom_right: Point2f = (top_left_x + dim_x, top_left_y + dim_y)
        bottom_left: Point2f = (top_left_x, top_left_y + dim_y)

        top_right = rotate_point(top_right, top_left, rotation)
        bottom_right = rotate_point(bottom_right, top_left, rotation)
        bottom_left = rotate_point(bottom_left, top_left, rotation)

        box = np.asarray(  # type: ignore[var-annotated]
            [top_left, top_right, bottom_right, bottom_left], dtype=np.int_
        )
        (center_x, center_y), (w, h), angle = cv2.minAreaRect(box)

        return BoundingBox(
            class_identifier=ClassIdentifier(
                class_id=class_id,
                class_name=class_name,
            ),
            box=Box.init_format_based(
                box_list=(float(center_x), float(center_y), float(w), float(h)),
                box_format=ObjectDetectionBBoxFormats.CXCYWH,
                angle=angle,
            ),
            score=1.0,
            difficult=False,
            occluded=False,
            content=content,
            background=False,
        )

    return BoundingBox(
        class_identifier=ClassIdentifier(
            class_id=class_id,
            class_name=class_name,
        ),
        box=Box.init_format_based(
            box_list=(top_left_x, top_left_y, dim_x, dim_y),
            box_format=ObjectDetectionBBoxFormats.XYWH,
            angle=0.0,
        ),
        score=1.0,
        difficult=False,
        occluded=False,
        content=content,
        background=False,
    )


def parse_segmentation(
    mapper: AnnotationClassMapper, label_studio_value: Dict[str, Any], image_shape: Tuple[int, int]
) -> Optional[Segmentation]:
    """
    Try to parse a single Segmentation object from a given label studio result value (one
    entry of a result in the annotation file).

    Args:
        mapper: The mapper that should be used to translate class names
        label_studio_value: The input value for parsing the BoundingBox
        image_shape: The shape of the related image

    Returns:
        Optional[Segmentation]: The parsed Segmentation object
    """
    polygonlabels = label_studio_value.get("polygonlabels", None)
    labels = label_studio_value.get("labels", None)
    if polygonlabels:
        label_studio_class_name = polygonlabels[0]
    elif labels:
        label_studio_class_name = labels[0]
    else:
        # We don't except annotations without a class
        return None

    try:
        class_name = mapper.map_annotation_class_name_to_model_class_name(
            class_name=label_studio_class_name
        )
        class_id = mapper.map_annotation_class_name_to_model_class_id(
            class_name=label_studio_class_name
        )
    except ClassMappingNotFoundError:
        logger.debug(
            "Could not find a valid class-mapping for class-name '%s'. "
            "Segmentation will be skipped",
            label_studio_class_name,
        )
        return None

    return Segmentation(
        class_identifier=ClassIdentifier(
            class_id=class_id,
            class_name=class_name,
        ),
        polygon=[
            [x / 100 * image_shape[1], y / 100 * image_shape[0]]
            for (x, y) in label_studio_value["points"]
        ],
        score=1.0,
        difficult=False,
        occluded=False,
        content=label_studio_value.get("text", [""])[0],
        background=False,
    )


def parse_from_annotation(
    results: Dict[Any, Any], mapper: AnnotationClassMapper
) -> Tuple[Optional[Tuple[int, int]], List[BoundingBox], List[Segmentation]]:

    logger.warning(
        "DEPRECATED: parse_from_annotation is deprecated and " "will be removed in future versions"
    )

    label_studio_annotation_data: LabelStudioAnnotationData = parse_annotation_data(
        results=results, mapper=mapper
    )

    return (
        label_studio_annotation_data.image_shape,
        label_studio_annotation_data.bounding_boxes,
        label_studio_annotation_data.segmentations,
    )


def parse_annotation_data(
    results: Dict[Any, Any], mapper: AnnotationClassMapper
) -> LabelStudioAnnotationData:
    """
    Parse all bounding-boxes and segmentations that are present in the results
    of a label studio annotation. If available parse the image-shape that should
    be associated with this annotation.

    Args:
        results: The input results for the parsing
        mapper: The mapper that should be used to translate class names

    Returns:
        Tuple[Optional[Tuple[int, int]], List[Classification], List[BoundingBox], List[Segmentation]]:
        The parsed data
    """
    image_shape: Optional[Tuple[int, int]] = None
    classifications: List[Classification] = []
    bounding_boxes: List[BoundingBox] = []
    segmentations: List[Segmentation] = []

    for result in results:
        result_types = result["type"]
        result_value = result["value"]

        try:
            image_shape = (int(result["original_height"]), int(result["original_width"]))
        except KeyError:
            # Classification tasks don't contain the entries
            # "original_height" and "original_width"
            image_shape = None

        if "choices" in result_types:
            classification: Optional[Classification] = parse_classification(
                mapper=mapper,
                label_studio_value=result_value,
            )
            if classification is not None:
                classifications.append(classification)

        if (
            "rectangle" in result_types or "rectanglelabels" in result_types
        ) and image_shape is not None:
            bounding_box: Optional[BoundingBox] = parse_bounding_box(
                mapper=mapper,
                label_studio_value=result_value,
                image_shape=image_shape,
            )
            if bounding_box is not None:
                bounding_boxes.append(bounding_box)
        elif (
            "polygon" in result_types or "polygonlabels" in result_types
        ) and image_shape is not None:
            segmentation: Optional[Segmentation] = parse_segmentation(
                mapper=mapper,
                label_studio_value=result_value,
                image_shape=image_shape,
            )
            if segmentation is not None:
                segmentations.append(segmentation)

    return LabelStudioAnnotationData(image_shape, classifications, bounding_boxes, segmentations)


class LabelStudioAnnotationParser(AnnotationParser):
    """
    Parser for the Label Studio json format of a json file that is
    generated by Label Studio per image

    Support data types:
    - BoundingBox
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        label_studio_input_data: List[AnnotationHandlerLabelStudioInputDataConfig],
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.label_studio_input_data = label_studio_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = []

        for dataset_count, input_data in enumerate(self.label_studio_input_data):
            image_file_dict = get_image_file_dict(
                input_dir=input_data.input_image_dir, image_format=input_data.image_format
            )

            label_studio_annotation_paths = [
                str(p) for p in Path(input_data.input_annotation_dir).glob("**/*") if p.is_file()
            ]

            for label_studio_annotation_path in label_studio_annotation_paths:
                label_studio_dict = get_label_studio_annotation_dict(
                    label_studio_annotation_path=label_studio_annotation_path
                )

                if label_studio_dict is None:
                    logger.warning(
                        "Label Studio annotation content could not be loaded from: '%s', "
                        "Annotation will be skipped." % label_studio_annotation_path
                    )
                    continue

                if not isinstance(label_studio_dict, dict):
                    logger.warning(
                        "Loaded Label Studio annotation content is not of type dict, file: '%s', "
                        "Annotation will be skipped." % label_studio_annotation_path
                    )
                    continue

                try:
                    label_studio_image_name = os.path.basename(
                        label_studio_dict["task"]["data"]["image"]
                    )

                    if label_studio_image_name in image_file_dict:
                        image_path = image_file_dict[label_studio_image_name]
                    else:
                        logger.warning(
                            "Label Studio annotation contains non-existing image: '%s', "
                            "Annotation will be skipped." % label_studio_image_name
                        )
                        continue

                    label_studio_annotation_data: LabelStudioAnnotationData = (
                        parse_annotation_data(
                            results=label_studio_dict["result"],
                            mapper=self.mapper,
                        )
                    )

                    if label_studio_annotation_data.image_shape is None:
                        _image_shape = FrameShape(*cv2.imread(image_path).shape)
                        label_studio_annotation_data.image_shape = (
                            _image_shape.height,
                            _image_shape.width,
                        )

                    annotations.append(
                        BaseAnnotation(
                            image_path=image_path,
                            annotation_path=label_studio_annotation_path,
                            image_shape=label_studio_annotation_data.image_shape,
                            classifications=label_studio_annotation_data.classifications,
                            bounding_boxes=label_studio_annotation_data.bounding_boxes,
                            segmentations=label_studio_annotation_data.segmentations,
                            image_dir=os.path.dirname(image_path),
                            annotation_dir=os.path.dirname(label_studio_annotation_path),
                            # TODO: How to set this?
                            replacement_string="",
                        )
                    )
                except (ValueError, ForbiddenClassError, KeyError) as error:
                    logger.exception(
                        msg="Could not parse annotation file '%s'. Annotation will be skipped."
                        % label_studio_annotation_path,
                        exc_info=error,
                    )

        return annotations



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for defining a component that enables to create registries for generic class types
"""

import importlib
import logging
from typing import Dict, Generic, Optional, TypeVar

from attr import define
from config_builder import BaseConfigClass
from related import StringField

ClassType = TypeVar("ClassType")

logger = logging.getLogger(__name__)


@define
class RegistryEntry(BaseConfigClass):
    """
    Basis class defining an entry in the registry.
    """

    module_type_name: str = StringField()
    module_constructor: str = StringField()
    package_name: str = StringField()


class MLCVZooRegistry(Generic[ClassType]):
    """
    Basis class that enables to create registries for generic class types
    """

    def __init__(self) -> None:
        self._registry: Dict[str, ClassType] = {}

    def register_module(
        self, module_type_name: str, module_constructor: ClassType, force: bool = False
    ) -> None:
        """
        Directly register a module constructor

        Args:
            module_type_name: Name of the model to register
            module_constructor: The reference to the model constructor
            force: Overwrite an existing entry

        Returns:
            None
        """

        if not force and module_type_name in self._registry:
            raise KeyError(
                f"{module_type_name} is already registered model registry in {self._registry}"
            )

        self._registry[module_type_name] = module_constructor

        logger.info(
            "Successfully registered module '%s' (%s)", module_type_name, module_constructor
        )

    def register_external_module(
        self,
        module_type_name: str,
        module_constructor: str,
        package_name: str,
        force: bool = False,
    ) -> Optional[str]:
        """
        Register an external model

        Args:
            module_type_name: Name of the model to register
            module_constructor: Name of the constructor of the model to register as string
            package_name: The full package to import to call the constructor
            force: Overwrite an existing entry

        Returns:
            None
        """
        try:
            # pylint: disable=c0415
            module = importlib.import_module(package_name)
            self.register_module(
                module_type_name=module_type_name,
                module_constructor=module.__dict__[module_constructor],
                force=force,
            )
            return module_type_name
        except KeyError as key_error:
            if "already registered model registry " in str(key_error):
                return module_type_name
        except ImportError as error:
            logger.error(
                "Optional module '%s' (%s.%s) not available: %s",
                module_type_name,
                package_name,
                module_constructor,
                error,
            )
        except Exception as error:  # pylint: disable=W0718
            logger.error(
                "Error while registering module '%s' (%s.%s): %s",
                module_type_name,
                package_name,
                module_constructor,
                error,
            )

        return None



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for parsing CVAT formatted annotations"""
import logging
import xml.etree.ElementTree as ET_xml
from typing import Any, Dict, List, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.api.data.types import Point2f
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError, ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerSingleFileInputDataConfig,
)
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.configuration.utils import str2bool
from mlcvzoo_base.data_preparation.annotation_builder.cvat_annotation_builder import (
    CVATAnnotationBuilder,
)

logger = logging.getLogger(__name__)


class CVATAnnotationParser(AnnotationParser):
    """
    AnnotationParser implementation for parsing CVAT .xml annotation files.

    Information about attributes:
    The CVAT tool allows to define specific attributes for each class that
    should be annotated. In Object Detection and Segmentation tasks, a well
    known attribute is "occluded". Since it is a very important attribute from
    the CVAT perspective, it has a specific functionality in its Web GUI
    for specifying this attribute. In the .xml annotation file it is part of the
    annotation item itself. Therefore, the CVATAnnotationParser will ignore
    the "occluded" attribute if it is specified via the additional attribute
    definition possibility of the CVAT tool.

    Example:
        <box label="car" source="manual" occluded="0" xtl="34.50" ytl="246.05" xbr="128.41" ybr="325.42" z_order="0">
          <attribute name="background">false</attribute>
          <attribute name="difficult">true</attribute>
          <attribute name="occluded">true</attribute>
        </box>

        In this case the "occluded" attribute that will be parsed in a mlcvzoo BoundinBox object
        will be set to false, since the item of the box annotation is set to "0".

    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        cvat_input_data: List[AnnotationHandlerSingleFileInputDataConfig],
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.cvat_input_data = cvat_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = []

        for dataset_count, input_data in enumerate(self.cvat_input_data):
            tree = ET_xml.parse(input_data.input_path)
            root = tree.getroot()

            if root is not None:
                images = root.findall("image")

                for image in images:
                    cvat_tags = image.findall("tag")
                    cvat_bboxes = image.findall("box")
                    cvat_polygons = image.findall("polygon")

                    try:
                        image_path = image.attrib["name"]
                        image_shape = (int(image.attrib["height"]), int(image.attrib["width"]))

                        classifications = self.__parse_cvat_classifications(cvat_tags=cvat_tags)
                        bounding_boxes = self.__parse_cvat_bounding_boxes(
                            cvat_bboxes=cvat_bboxes,
                            use_difficult=input_data.use_difficult,
                            use_background=input_data.use_background,
                            use_occluded=input_data.use_occluded,
                            image_shape=image_shape,
                        )
                        segmentations = self.__parse_cvat_polygons(
                            cvat_polygons=cvat_polygons,
                            use_difficult=input_data.use_difficult,
                            use_background=input_data.use_background,
                            use_occluded=input_data.use_occluded,
                        )

                        replacement_string = (
                            AnnotationParser.csv_directory_replacement_string.format(dataset_count)
                        )

                        cvat_builder = CVATAnnotationBuilder(
                            image_shape=image_shape,
                            input_classifications=classifications,
                            input_bounding_boxes=bounding_boxes,
                            input_segmentations=segmentations,
                        )

                        annotation = cvat_builder.build(
                            image_path=image_path,
                            annotation_path=input_data.input_path,
                            image_dir=input_data.input_root_dir,
                            annotation_dir=input_data.input_root_dir,
                            replacement_string=replacement_string,
                        )
                        annotations.append(annotation)

                    except (ValueError, ForbiddenClassError) as error:
                        logger.warning("%s, annotation will be skipped" % str(error))
                        continue

        return annotations

    def __parse_cvat_classifications(
        self, cvat_tags: List[ET_xml.Element]
    ) -> List[Classification]:
        classifications: List[Classification] = []

        for tag in cvat_tags:
            xml_class_name = tag.attrib["label"]

            try:
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=xml_class_name
                )
                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=xml_class_name
                )
            except ClassMappingNotFoundError:
                logger.debug(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "Classification will be skipped",
                    xml_class_name,
                )
                continue

            classification = Classification(
                class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                model_class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                score=1.0,
            )

            classifications.append(classification)

        return classifications

    def __parse_cvat_bounding_boxes(
        self,
        cvat_bboxes: List[ET_xml.Element],
        use_difficult: bool,
        use_background: bool,
        use_occluded: bool,
        image_shape: Tuple[int, int],
    ) -> List[BoundingBox]:
        bounding_boxes: List[BoundingBox] = []

        for cvat_bbox in cvat_bboxes:
            xml_class_name = cvat_bbox.attrib["label"]

            try:
                # map the parsed "class_name" according to the mapping defined in the mapper class
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=xml_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=xml_class_name
                )
            except ClassMappingNotFoundError:
                logger.debug(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "BndBox will be skipped, cvat-box= '%s'",
                    xml_class_name,
                    cvat_bbox,
                )
                continue

            try:
                (
                    difficult,
                    background,
                    content,
                    meta_attributes,
                ) = CVATAnnotationParser.__read_attributes(cvat_bbox.findall("attribute"))

                occluded = str2bool(
                    [item for item in cvat_bbox.items() if item[0] == "occluded"][0][1]
                )
                bounding_box = BoundingBox(
                    class_identifier=ClassIdentifier(
                        class_id=class_id,
                        class_name=class_name,
                    ),
                    model_class_identifier=ClassIdentifier(
                        class_id=class_id,
                        class_name=class_name,
                    ),
                    score=1.0,
                    box=Box.init_format_based(
                        box_list=(
                            float(cvat_bbox.attrib["xtl"]),
                            float(cvat_bbox.attrib["ytl"]),
                            float(cvat_bbox.attrib["xbr"]),
                            float(cvat_bbox.attrib["ybr"]),
                        ),
                        box_format=ObjectDetectionBBoxFormats.XYXY,
                        src_shape=image_shape,
                    ),
                    occluded=occluded,
                    difficult=difficult,
                    background=background,
                    content=content,
                    meta_attributes=meta_attributes,
                )
            except (ValueError, IndexError) as error:
                logger.warning("%s, Bounding-Box will be skipped" % str(error))
                continue

            if difficult and not use_difficult:
                continue
            if occluded and not use_occluded:
                continue
            if background and not use_background:
                continue

            bounding_boxes.append(bounding_box)

        return bounding_boxes

    def __parse_cvat_polygons(
        self,
        cvat_polygons: List[ET_xml.Element],
        use_difficult: bool,
        use_background: bool,
        use_occluded: bool,
    ) -> List[Segmentation]:
        segmentations: List[Segmentation] = []

        for cvat_polygon in cvat_polygons:
            xml_class_name = cvat_polygon.attrib["label"]

            try:
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=xml_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=xml_class_name
                )
            except ClassMappingNotFoundError:
                logger.debug(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "Segmentation will be skipped, cvat-segmentation= '%s'",
                    xml_class_name,
                    cvat_polygon,
                )
                continue

            cvat_points = cvat_polygon.attrib["points"].split(";")

            polygon: List[Point2f] = []

            for cvat_point in cvat_points:
                point_split = cvat_point.split(",")
                polygon.append([float(point_split[0]), float(point_split[1])])

            (
                difficult,
                background,
                content,
                meta_attributes,
            ) = CVATAnnotationParser.__read_attributes(cvat_polygon.findall("attribute"))

            occluded = str2bool(
                [item for item in cvat_polygon.items() if item[0] == "occluded"][0][1]
            )
            segmentation = Segmentation(
                class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                model_class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                score=1.0,
                polygon=polygon,
                occluded=occluded,
                difficult=difficult,
                background=background,
                content=content,
                meta_attributes=meta_attributes,
            )

            if difficult and not use_difficult:
                continue

            if occluded and not use_occluded:
                continue

            if background and not use_background:
                continue

            segmentations.append(segmentation)

        return segmentations

    @staticmethod
    def __read_attributes(
        attributes: List[ET_xml.Element],
    ) -> Tuple[bool, bool, str, Dict[str, Any]]:
        difficult = False
        background = False
        content = ""
        meta_attributes: Dict[str, Any] = {}

        for attribute in attributes:
            if attribute.attrib["name"] == "difficult":
                difficult = False if attribute.text == "false" else True
            elif attribute.attrib["name"] == "background":
                background = False if attribute.text == "false" else True
            elif attribute.attrib["name"] == "content":
                if attribute.text is not None:
                    content = str(attribute.text)
            elif attribute.attrib["name"] == "occluded":
                # The occluded attribute is handled via an extra field in the CVAT xml and
                # not via a manual configured CVAT attribute
                pass
            else:
                if attribute.attrib["name"] is not None and attribute.text is not None:
                    meta_attributes[attribute.attrib["name"]] = attribute.text

        return difficult, background, content, meta_attributes

    def parse_cvat_meta_info(self) -> List[ET_xml.Element]:
        """
        Parses all tags under meta tag into a list of XML Elements.

        Returns: a list of XML Elements

        """

        meta_elements: List[ET_xml.Element] = []
        for dataset_count, input_data in enumerate(self.cvat_input_data):
            tree = ET_xml.parse(input_data.input_path)
            root = tree.getroot()

            if root is not None:
                meta_tags = root.findall("meta")

                if len(meta_tags) > 1:
                    logger.warning(
                        "Found more than one <meta> tag in given XML file. "
                        "Format is not as expected."
                    )
                meta_elements.append(meta_tags[0])

        return meta_elements



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for configuring the parsing of information from yaml in python
accessible attributes for the ReadFromFileModel class
"""
from typing import Optional

import related
from attr import define

from mlcvzoo_base.api.configuration import ModelConfiguration
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.configuration.reduction_mapping_config import ReductionMappingConfig


@define
class ReadFromFileConfig(ModelConfiguration):
    """
    Class for parsing information from yaml in respective hierarchy

    Attributes:
        class_mapping (ClassMappingConfig):
        annotation_handler_config:
        use_image_name_hash:
        include_segmentations:
        reduction_class_mapping:
    """

    class_mapping: ClassMappingConfig = related.ChildField(ClassMappingConfig)

    annotation_handler_config: AnnotationHandlerConfig = related.ChildField(
        cls=AnnotationHandlerConfig
    )

    use_image_name_hash: bool = related.BooleanField(default=False)

    include_segmentations: bool = related.BooleanField(default=False)

    reduction_class_mapping: Optional[ReductionMappingConfig] = related.ChildField(
        cls=ReductionMappingConfig, required=False, default=None
    )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for storing types that are shared across the mlcvzoo
"""

from __future__ import annotations

from dataclasses import dataclass
from math import isclose
from typing import List, NamedTuple, Sequence, Union

import numpy as np

from mlcvzoo_base.api.structs import float_equality_precision

try:
    from typing import Literal

    from numpy.typing import NDArray
    from typing_extensions import Annotated

    ImageType = Annotated[NDArray[np.int_], Literal["Height", "Width", 3]]
    PolygonTypeNP = Annotated[NDArray[np.float_], Literal["Length", 2]]
    Point2fNP = Annotated[NDArray[np.float_], Literal[1, 2]]
    Point2DNP = Annotated[NDArray[np.int_], Literal[1, 2]]
except ImportError as import_error:
    # NDArray is available from numpy>=1.21.0
    ImageType = np.ndarray  # type: ignore[misc]
    PolygonTypeNP = np.ndarray  # type: ignore[misc]
    Point2fNP = np.ndarray  # type: ignore[misc]
    Point2DNP = np.ndarray  # type: ignore[misc]


# Point as [x, y]
Point2f = Sequence[float]
Point2D = Sequence[int]
PolygonType = List[Point2f]


class FrameShape(NamedTuple):
    height: int
    width: int
    channel: int = 3


@dataclass
class MetricInfo:
    """
    Dataclass for storing information about a checkpoint and its metric
    """

    path: str
    epoch: int
    name: str
    score: float


def point_equal(point: Point2f, other_point: Point2f) -> bool:
    """
    Args:
        point: The first point to compare
        other_point: The second point to compare

    Returns:
         Whether the two given points are equal
    """
    return isclose(a=point[0], b=other_point[0], abs_tol=float_equality_precision) and isclose(
        a=point[1], b=other_point[1], abs_tol=float_equality_precision
    )


def polygon_equal(
    polygon: Union[PolygonType, PolygonTypeNP], other_polygon: Union[PolygonType, PolygonTypeNP]
) -> bool:
    """
    Args:
        polygon: The first polygon to compare
        other_polygon: The second polygon to compare

    Returns:
         Whether the two given polygons are equal
    """
    if len(polygon) == len(other_polygon):
        for polygon_point, other_polygon_point in zip(polygon, other_polygon):
            if not point_equal(point=polygon_point, other_point=other_polygon_point):
                return False
    else:
        return False

    return True



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Data class that holds attributes of Classification objects"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from mlcvzoo_base.api.data.box import Box, GeometricPerception
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.types import PolygonTypeNP


class GeometricClassification(GeometricPerception, ABC, Classification):
    def __init__(
        self,
        class_identifier: ClassIdentifier,
        score: float,
        model_class_identifier: Optional[ClassIdentifier] = None,
    ) -> None:
        Classification.__init__(
            self,
            class_identifier=class_identifier,
            model_class_identifier=model_class_identifier,
            score=score,
        )

    @abstractmethod
    def box(self) -> Box:
        raise NotImplementedError(f"Call of .box() from abstract class {self.__class__}")

    @abstractmethod
    def ortho_box(self) -> Box:
        raise NotImplementedError(f"Call of .ortho_box() from abstract class {self.__class__}")

    @abstractmethod
    def polygon(self) -> PolygonTypeNP:
        raise NotImplementedError(f"Call of .polygon() from abstract class {self.__class__}")
